# -*- coding: utf-8 -*-
"""
-------------------------------
(c) Universite catholique de Louvain, 2019

Author: Robotran Team
Last update : 2020
-------------------------------
"""
import os
import sys
import numpy
import ctypes
ndpointer = numpy.ctypeslib.ndpointer
import platform
import warnings
warnings.simplefilter('always', ImportWarning)

#==============================================================================
# Global parameter of the current module
#==============================================================================
__MODULE_DIR__ = os.path.dirname(os.path.abspath(__file__))
__OS__ = platform.uname()[0]

#==============================================================================
# Path to the libraries
#==============================================================================
try:
    from .libpath import __LIB_PATH__
    if  __OS__ == "Windows":
        __LIB_PATH__ = os.path.join(__LIB_PATH__,'Release')
except ImportError:
    __LIB_PATH__ = os.path.join(__MODULE_DIR__,'libs64')
    if __OS__ == "Linux":
        __LIB_PATH__ = os.path.join(__LIB_PATH__,'linux')
    elif __OS__ == "Darwin":
        __LIB_PATH__ = os.path.join(__LIB_PATH__,'darwin')
    elif __OS__ == "Windows":
        __LIB_PATH__ = os.path.join(__LIB_PATH__,'windows')

#==============================================================================
# Loading libraries
#==============================================================================
if __OS__ == "Linux":
    ctypes.DEFAULT_MODE = ctypes.RTLD_GLOBAL
elif not(__OS__ in ["Darwin", "Windows"]):
    raise ImportError("Unsupported Operating System identifier: "+__OS__)

# Load Tire Ground Car (TGC) library
lib_tgc = numpy.ctypeslib.load_library("libAddon_TireGroundCar",__LIB_PATH__)

#==============================================================================
# Pre-declaration of required structures (Python class)
#==============================================================================
# None are required, the builded C API uses void*

#==============================================================================
# Filling required structure (Python class)
#==============================================================================
# None are required

#==============================================================================
# Define functions arguments and return type
#==============================================================================
lib_tgc.tgc_car_kine_wheel.argtypes =[ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"), 
                                      ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"), 
                                      ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"), 
                                      ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"), 
                                      ctypes.c_double,
                                      ctypes.POINTER(ctypes.c_double),
                                      ctypes.POINTER(ctypes.c_double),
                                      ctypes.POINTER(ctypes.c_double),
                                      ctypes.POINTER(ctypes.c_double),
                                      ctypes.POINTER(ctypes.c_double),
                                      ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"), 
                                      ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"), 
                                      ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"), 
                                      ndpointer(ctypes.c_double, flags="C_CONTIGUOUS")
                                      ]
lib_tgc.tgc_car_kine_wheel.restype  = None

lib_tgc.tgc_bakker_contact.argtypes =[ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                                      ndpointer(ctypes.c_double, flags="C_CONTIGUOUS"),
                                      ctypes.c_double, ctypes.c_double, ctypes.c_double
                                      ]
lib_tgc.tgc_bakker_contact.restype  = None

#==============================================================================
# Define exported variables
#==============================================================================
# None are exported

#==============================================================================
# Define package/module functions
#==============================================================================
def tgc_car_kine_wheel(Pw, Rw, Vw, OMw, rnom):
    """
    Compute car wheel kinematics which are necessary inputs for a tire/ground contact model.
    
    Note that the car tire is defined by a nominal radius (from wheel center to
    contact patch).
    
    Be advised: the geometrical contact point velocity can NOT be zero : the slip would be infinite
    This function should only be called for a tyre in motion. 
    
    This function calls a C-library.
    
    Parameters
    ----------
    Pw : numpy.ndarray
         Position of the wheel center in the inertial frame as [3., P_x, P_y, P_z].
    Rw : numpy.ndarray
         Rotation Matrix of the wheel body of shape (4,4), the rotation matrix
         is contained in Rw[1:,1:].
    Vw : numpy.ndarray
         Velocity of the wheel center in the inertial frame as [3., V_x, V_y, V_z].
    OMw : numpy.ndarray
          Angular velocity of the wheel center in the inertial frame [3., W_x, W_y, W_z].
    rnom : float
           Wheel nominal radius.
   
    Returns
    -------
    pen : float
          Tire vertical deflection.
    rz : float
         Distance between the wheel center and the intersection of frame [S] with
         the flat ground.
    angslip : float
              Lateral slip angle.
    angcamb : float
              Camber angle.
    slip : float
           Longitudinal slip. Be careful that the geometrical contact point velocity can not be zero (infinite slip !).
    Pct : numpy.ndarray
          Position of the contact point expressed in the inertial frame as 
          [3., Pct_x, Pct_y, Pct_z].
    Vmct : numpy.ndarray
           Velocity of the material point in contact expressed in the inertial 
           frame as [3., Vmct_x, Vmct_y, Vmct_z].
    Rt_ground : numpy.ndarray
                Transpose of Rotation matrix from [I] to [T] of shape (4,4), the
                rotation matrix is contained in Rt_ground[1:,1:].
    dxF : numpy.ndarray
          Vector of force application (from the wheel center to the contact point)
          expressed in the rotating wheel frame [Y] as [3., dxF_x, dxF_y, dxF_z]
    """
    # prepare results variables
    pen = ctypes.c_double(0.0)
    rz = ctypes.c_double(0.0)
    angslip = ctypes.c_double(0.0)
    angcamb = ctypes.c_double(0.0)
    slip = ctypes.c_double(0.0)
    Pcontact = numpy.zeros(4)
    Pcontact[0] = 3.0
    Vcontact = numpy.zeros(4)
    Vcontact[0] = 3.0
    Rt_ground = numpy.zeros(16)
    Rt_ground[0] = 3.0
    Rt_ground[4] = 3.0
    Rt_ground[8] = 3.0
    Rt_ground[12] = 3.0
    dxF = numpy.zeros(4)
    dxF[0] = 3.0
    
    lib_tgc.tgc_car_kine_wheel(Pw, Rw, Vw, OMw, rnom,
                               ctypes.byref(pen), ctypes.byref(rz), ctypes.byref(angslip), 
                               ctypes.byref(angcamb), ctypes.byref(slip),
                               Pcontact, Vcontact, Rt_ground, dxF
                               )
    
    return (pen.value, rz.value, angslip.value, angcamb.value, slip.value, 
            Pcontact, Vcontact, Rt_ground, dxF)

def tgc_bakker_contact(Fwhl, Mwhl, angslip, angcamb, slip):
    """
    Compute tire/ground contact forces according to the Bakker model.
    
    Parameters
    ----------
    Fwhl : numpy.ndarray
           Tire/ground contact forces expressed in [R] frame, will be filled 
           with Fwhl[1:4] = [Fwhl_x, Fwhl_y, Fwhl_z].
    Mwhl : numpy.ndarray
           tire/ground contact torques expressed in [R] frame, will be filled 
           with Mwhl[1:4] = [Mwhl_x, Mwhl_y, Mwhl_z].
    angslip : float
              Lateral slip angle expressed in rad.
    angcamb : float
              Camber angle expressed in rad.
    slip : float
           Longitudinal slip. Be careful that the geometrical contact point velocity can not be zero (infinite slip !).
    Return
    ------
    Nothing, computation results are stored in the arguments
    """
    lib_tgc.tgc_bakker_contact(Fwhl, Mwhl, angslip, angcamb, slip)
   
   
